.sphx-glr-thumbcontainer {
    background: inherit !important;
    min-height: 250px !important;
    margin: 10px !important;
}

.sphx-glr-thumbcontainer .headerlink {
    display: none !important;
}

div.sphx-glr-thumbcontainer span {
    font-style: normal !important;
}

.sphx-glr-thumbcontainer a.internal {
    padding: 140px 10px 0 !important;
}

.sphx-glr-thumbcontainer .figure {
    width: 200px !important;
}

.sphx-glr-thumbcontainer .figure.align-center {
    text-align: center;
    margin-left: 0%;
    transform: translate(0%);
}
